/*  
    Conditional Temporal Median Filter

    Version 0.90  June 27, 2003

    Copyright (C) 2003 Kevin Atkinson (kevin.med@atkinson.dhs.org) under 
    the GNU GPL version 2.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <cstdlib>
#include <cstring>

using std::memset;

#include "med.hpp"

extern "C"
void memzero(void * d0, size_t s)
{
  memset(d0, 0, s);
  //int * d = (int *)d0;
  //for (size_t i = 0; i != s/4; ++i)
  //  *d = 0;
}

//
// Motion Detection
//

extern "C"
void find_motion(byte * p, byte * c, byte * motion, size_t size)
{
  for (size_t i = 0; i != size; ++i) {
    if (abs(p[i] - c[i]) > MTHRES) motion[i] = 3;
    else                           motion[i] = 0;
  }
}

extern "C"
void motion_denoise(byte * moving, int w, int h)
{
  // This function is based on the motion map denoise code found in
  // Donald Graft's Smart Deinterlacer Filter.  I completely rewrote
  // it to so that its complexity is O(w*h) instead of
  // O(w*h*MWIDTH*MHEIGHT).

  byte fmoving[h*w];
  byte row[Y_NEXT][((w+MWIDTH)/8)*8 + 8];

  const int MHEIGHT_2 = MHEIGHT/2;
  const int MWIDTH_2  = MWIDTH/2;

  // Erode
  memzero(row, sizeof(row));
  for (int y = -(MHEIGHT_2)*Y_NEXT; y < h; ++y)
  {
    byte * r = row[y & (Y_NEXT - 1)] + MWIDTH_2 + 1;
    int y1 = y - (MHEIGHT_2)*Y_NEXT - Y_NEXT, y2 = y + (MHEIGHT_2)*Y_NEXT;
    if (y1 >= 0) {
      byte * r2 = moving + w * y1;
      for (int x = 0; x != w; ++x)
        r[x] -= r2[x];}
    if (y2 < h) {
      byte * r2 = moving + w * y2;
      for (int x = 0; x != w; ++x)
        r[x] += r2[x];}
    if (y >= 0) {
      byte * r2 = moving + w * y;
      byte * fw = fmoving + w * y;
      int total = 0;
      for (int x = 0; x != MWIDTH_2; ++x)
        total += r[x];
      for (int x = 0; x != w; ++x) {
        total -= r[x - MWIDTH_2 - 1];
        total += r[x + MWIDTH_2];
        fw[x] = r2[x] && total > MP*3 ? 1 : 0;}}
  }

  // Dilute
  memzero(row, sizeof(row));
  for (int y = -(MHEIGHT_2)*Y_NEXT; y < h; ++y)
  {
    byte * r = row[y & (Y_NEXT - 1)] + MWIDTH_2 + 1;
    int y1 = y - (MHEIGHT_2)*Y_NEXT - Y_NEXT, y2 = y + (MHEIGHT_2)*Y_NEXT;
    if (y1 >= 0) {
      byte * r2 = fmoving + w * y1;
      for (int x = 0; x != w; ++x)
        r[x] -= r2[x];}
    if (y2 < h) {
      byte * r2 = fmoving + w * y2;
      for (int x = 0; x != w; ++x)
        r[x] += r2[x];}
    if (y >= 0) {
      byte * fw = moving + w * y;
      int total = 0;
      for (int x = 0; x != MWIDTH_2; ++x)
        total += r[x];
      for (int x = 0; x != w; ++x) {
        total -= r[x - MWIDTH_2 - 1];
        total += r[x + MWIDTH_2];
        fw[x] = total ? 255 : 0;}}
  }
}

//
// Noise Detection
//

extern "C"
void cond_median(byte * p, 
                 byte * c, byte * c_noise, byte * c_motion, 
                 byte * n, byte * n_motion, 
                 byte * o, size_t size)
{
  for (size_t i = 0; i != size; ++i) {
    o[i] = c[i];
    if (c_motion[i] || n_motion[i]) continue;
    byte mn = min(p[i], n[i]);
    byte mx = max(p[i], n[i]);
    o[i] = max(mn, o[i]);
    o[i] = min(mx, o[i]);
  }
}

extern "C"
void find_outliers(byte * p, byte * c, byte * n, byte * o, 
                   size_t size)
{
  for (size_t i = 0; i != size; ++i) {
    o[i] = 0;
    byte mn = min(p[i], n[i]);
    byte mx = max(p[i], n[i]);
    if (c[i] < mn) o[i] = mn - c[i];
    if (c[i] > mx) o[i] = c[i] - mx;
  }
}


extern "C"
void mark_motion(byte * p, byte * p_noise,
                 byte * c, byte * c_noise, byte * c_motion,
                 size_t size)
{
  for (size_t i = 0; i != size; ++i) {
    if (abs(p[i] - c[i]) > MTHRES) {
      if (p_noise[i] || c_noise[i])  c_motion[i] = 1;
      else                           c_motion[i] = 3;
    } else                           c_motion[i] = 0;
  }
}

//
//
//

extern "C"
void remove_outliers(byte * p,
                     byte * c, byte * c_noise, byte * c_motion,
                     byte * n, byte * n_motion,
                     byte * o, 
                     size_t size)
{
  for (size_t i = 0; i != size; ++i)
  {
    if (c_noise[i] && !c_motion[i] && !n_motion[i])
      o[i] = (p[i] + n[i] + 1)/2;
    else
      o[i] = c[i];
  }
}

extern "C"
void mark_outliers(byte * p,
                   byte * c, byte * c_noise, byte * c_motion,
                   byte * n, byte * n_motion,
                   byte * o, 
                   size_t size)
{
  for (size_t i = 0; i != size; ++i)
  {
    if (c_noise[i] && !(c_motion[i] || n_motion[i]))
      o[i] = MARK_V;
    else
      o[i] = c[i];
  }
}

extern "C"
void map_outliers(byte * p,
                  byte * c, byte * c_noise, byte * c_motion,
                  byte * n, byte * n_motion,
                  byte * o, 
                  size_t size)
{
  for (size_t i = 0; i != size; ++i)
  {
    if (c_noise[i]) {
      if (c_motion[i] || n_motion[i])
        o[i] = 159;
      else
        o[i] = 255;
    } else {
      if (c_motion[i]) 
        o[i] = 95;
      else if (n_motion[i]) 
        o[i] = 63;
      else
        o[i] = 0;
    }
  }
}
