/*  
    Conditional Temporal Median Filter

    Version 0.90  June 27, 2003

    Copyright (C) 2003 Kevin Atkinson (kevin.med@atkinson.dhs.org) under 
    the GNU GPL version 2.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <cstdlib>
#include <cstring>

using std::memset;

#include "sse.hpp"
#include "med.hpp"

extern "C"
void memzero(void * d0, size_t s)
{
  memset(d0, 0, s);
  //int * d = (int *)d0;
  //for (size_t i = 0; i != s/4; ++i)
  //  *d = 0;
}

//
// Motion Detection
//

extern "C"
void find_motion(v8ubyte * p, v8ubyte * c, v8ubyte * motion, size_t size)
{
  size /= 8;
  v8ubyte thres; thres.set(MTHRES+1);
  v8ubyte v_3; v_3.set(3);
  for (size_t i = 0; i != size; ++i) 
  {
    v8ubyte mask = dist(p[i], c[i]) >= thres;
    motion[i] = mask & v_3;
  }
  mm_empty();
}

extern "C"
void motion_denoise(byte * moving, int w, int h)
{
  // This function is based on the motion map denoise code found in
  // Donald Graft's Smart Deinterlacer Filter.  I completely rewrote
  // it to so that its complexity is O(w*h) instead of
  // O(w*h*MWIDTH*MHEIGHT).

  byte fmoving[h*w];
  byte row[Y_NEXT][((w+64)/8)*8 + 8];
  int w_8 = w/8;

  const int MHEIGHT_2 = MHEIGHT/2;
  const int MWIDTH_2  = MWIDTH/2;

  // Erode
  memzero(row, sizeof(row));
  for (int y = -(MHEIGHT/2)*Y_NEXT; y < h; ++y)
  {
    byte * r = row[y & (Y_NEXT - 1)] + 32;
    v8ubyte * r_v = (v8ubyte *)r;
    int y1 = y - (MHEIGHT_2)*Y_NEXT - Y_NEXT, y2 = y + (MHEIGHT_2)*Y_NEXT;
    if (y1 >= 0) {
      v8ubyte * r2_v = (v8ubyte *)(moving + w * y1);
      for (int x = 0; x != w_8; ++x)
        r_v[x] -= r2_v[x];}
    if (y2 < h) {
      v8ubyte * r2_v = (v8ubyte *)(moving + w * y2);
      for (int x = 0; x != w_8; ++x)
        r_v[x] += r2_v[x];}
    if (y >= 0) {
      byte * r2 = moving + w * y;
      byte * fw = fmoving + w * y;
      int total = 0;
      for (int x = 0; x != MWIDTH_2; ++x)
        total += r[x];
      for (int x = 0; x != w; ++x) {
        total -= r[x - MWIDTH_2 - 1];
        total += r[x + MWIDTH_2];
        fw[x] = r2[x] && total > MP*3 ? 1 : 0;}}
  }

  // Dilute
  memzero(row, sizeof(row));
  for (int y = -(MHEIGHT_2)*Y_NEXT; y < h; ++y)
  {
    byte * r = row[y & (Y_NEXT - 1)] + 32;
    v8ubyte * r_v = (v8ubyte *)r;
    int y1 = y - (MHEIGHT_2)*Y_NEXT - Y_NEXT, y2 = y + (MHEIGHT_2)*Y_NEXT;
    if (y1 >= 0) {
      v8ubyte * r2_v = (v8ubyte *)(fmoving + w * y1);
      for (int x = 0; x != w_8; ++x)
        r_v[x] -= r2_v[x];}
    if (y2 < h) {
      v8ubyte * r2_v = (v8ubyte *)(fmoving + w * y2);
      for (int x = 0; x != w_8; ++x)
        r_v[x] += r2_v[x];}
    if (y >= 0) {
      byte * fw = moving + w * y;
      int total = 0;
      for (int x = 0; x != MWIDTH_2; ++x)
        total += r[x];
      for (int x = 0; x != w; ++x) {
        total -= r[x - MWIDTH_2 - 1];
        total += r[x + MWIDTH_2];
        fw[x] = total ? 255 : 0;}}
  }
  mm_empty();
}

//
// Noise Detection
//

extern "C"
void cond_median(v8ubyte * p,
                 v8ubyte * c, v8ubyte * c_noise, v8ubyte * c_motion, 
                 v8ubyte * n, v8ubyte * n_motion,
                 v8ubyte * o, size_t size)
{
  size /= 8;
  for (size_t i = 0; i != size; ++i) {
    v8ubyte mask = c_motion[i] | n_motion[i];
    v8ubyte mn = min(p[i], n[i]);
    v8ubyte mx = max(p[i], n[i]);
    v8ubyte tmp = c[i];
    tmp = max(tmp, mn);
    tmp = min(tmp, mx);
    o[i] = if_else(mask, c[i], tmp);
  }
  mm_empty();
}

extern "C"
void find_outliers(v8ubyte * p, v8ubyte * c, v8ubyte * n, v8ubyte * o, 
                   size_t size)
{
  size /= 8;
  for (size_t i = 0; i != size; ++i) {
    v8ubyte mn = min(p[i], n[i]);
    v8ubyte mx = max(p[i], n[i]);
    o[i] = ssub(mn, c[i]) | ssub(c[i], mx);
  }
  mm_empty();
}

extern "C"
void mark_motion(v8ubyte * p, v8ubyte * p_noise,
                 v8ubyte * c, v8ubyte * c_noise, v8ubyte * c_motion,
                 size_t size)
{
  size /= 8;
  v8ubyte thres; thres.set(MTHRES+1);
  v8ubyte v_1; v_1.set(1);
  v8ubyte v_3; v_3.set(3);
  v8ubyte m1,m2,m3,m4;
  for (size_t i = 0; i != size; ++i)
  {
    v8ubyte d_mask = dist(p[i], c[i]) >= thres;
    v8ubyte n_mask = p_noise[i] | c_noise[i];
    c_motion[i] = d_mask & if_else(n_mask, v_1, v_3);
  }
  mm_empty();
}

//
//
//

// this may have horrible cache performance
extern "C"
void remove_outliers(v8ubyte * p,
                     v8ubyte * c, v8ubyte * c_noise, v8ubyte * c_motion,
                     v8ubyte * n, v8ubyte * n_motion,
                     v8ubyte * o, 
                     size_t size)
{
  size /= 8;
  for (size_t i = 0; i != size; ++i)
  {
    v8ubyte mask = andn(c_motion[i] | n_motion[i], c_noise[i]);
    o[i] = if_else(mask, avg(p[i], n[i]), c[i]);
  }
  mm_empty();
}

//mark_outliers can be MMX/SSE optimized
extern "C"
void mark_outliers(byte * p,
                   byte * c, byte * c_noise, byte * c_motion,
                   byte * n, byte * n_motion,
                   byte * o, 
                   size_t size)
{
  for (size_t i = 0; i != size; ++i)
  {
    if (c_noise[i] && !(c_motion[i] || n_motion[i]))
      o[i] = MARK_V;
    else
      o[i] = c[i];
  }
}

//map_outliers can be MMX/SSE optimized
extern "C"
void map_outliers(byte * p,
                  byte * c, byte * c_noise, byte * c_motion,
                  byte * n, byte * n_motion,
                  byte * o, 
                  size_t size)
{
  for (size_t i = 0; i != size; ++i)
  {
    if (c_noise[i]) {
      if (c_motion[i] || n_motion[i])
        o[i] = 159;
      else
        o[i] = 255;
    } else {
      if (c_motion[i]) 
        o[i] = 95;
      else if (n_motion[i]) 
        o[i] = 63;
      else
        o[i] = 0;
    }
  }
}
